{
  "type": "record",
  "name": "EntityChangeEvent",
  "namespace": "com.linkedin.pegasus2avro.platform.event.v1",
  "doc": "Shared fields for all entity change events.",
  "fields": [
    {
      "name": "entityType",
      "type": "string",
      "doc": "The type of the entity affected. Corresponds to the entity registry, e.g. 'dataset', 'chart', 'dashboard', etc."
    },
    {
      "name": "entityUrn",
      "type": "string",
      "doc": "The urn of the entity which was affected.",
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    },
    {
      "name": "category",
      "type": "string",
      "doc": "The category type (TAG, GLOSSARY_TERM, OWNERSHIP, TECHNICAL_SCHEMA, etc). This is used to determine what the rest of the schema will look like."
    },
    {
      "name": "operation",
      "type": "string",
      "doc": "The operation type. This is used to determine what the rest of the schema will look like."
    },
    {
      "name": "modifier",
      "type": [
        "null",
        "string"
      ],
      "doc": "The urn of the entity which was affected.",
      "default": null
    },
    {
      "name": "parameters",
      "type": [
        "null",
        {
          "type": "record",
          "name": "Parameters",
          "doc": "Arbitrary key-value parameters for an Entity Change Event. (any record).",
          "fields": []
        }
      ],
      "doc": "Arbitrary key-value parameters corresponding to the event.",
      "default": null
    },
    {
      "name": "auditStamp",
      "type": {
        "type": "record",
        "name": "AuditStamp",
        "namespace": "com.linkedin.pegasus2avro.common",
        "doc": "Data captured on a resource/association/sub-resource level giving insight into when that resource/association/sub-resource moved into a particular lifecycle stage, and who acted to move it into that specific lifecycle stage.",
        "fields": [
          {
            "name": "time",
            "type": "long",
            "doc": "When did the resource/association/sub-resource move into the specific lifecycle stage represented by this AuditEvent."
          },
          {
            "name": "actor",
            "type": "string",
            "doc": "The entity (e.g. a member URN) which will be credited for moving the resource/association/sub-resource into the specific lifecycle stage. It is also the one used to authorize the change.",
            "java": {
              "class": "com.linkedin.pegasus2avro.common.urn.Urn"
            }
          },
          {
            "name": "impersonator",
            "type": [
              "null",
              "string"
            ],
            "doc": "The entity (e.g. a service URN) which performs the change on behalf of the Actor and must be authorized to act as the Actor.",
            "default": null,
            "java": {
              "class": "com.linkedin.pegasus2avro.common.urn.Urn"
            }
          }
        ]
      },
      "doc": "Audit stamp of the operation"
    },
    {
      "name": "version",
      "type": "int",
      "doc": "The version of the event type, incremented in integers."
    }
  ],
  "Event": {
    "name": "entityChangeEvent"
  }
}