{
  "type": "record",
  "name": "CorpGroupEditableInfo",
  "namespace": "com.linkedin.pegasus2avro.identity",
  "doc": "Group information that can be edited from UI",
  "fields": [
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "A description of the group",
      "default": null,
      "Searchable": {
        "fieldName": "editedDescription",
        "fieldType": "TEXT"
      }
    },
    {
      "name": "pictureLink",
      "type": "string",
      "doc": "A URL which points to a picture which user wants to set as the photo for the group",
      "default": "https://raw.githubusercontent.com/datahub-project/datahub/master/datahub-web-react/src/images/default_avatar.png",
      "java": {
        "class": "com.linkedin.pegasus2avro.common.url.Url",
        "coercerClass": "com.linkedin.pegasus2avro.common.url.UrlCoercer"
      }
    },
    {
      "name": "slack",
      "type": [
        "null",
        "string"
      ],
      "doc": "Slack channel for the group",
      "default": null
    },
    {
      "name": "email",
      "type": [
        "null",
        "string"
      ],
      "doc": "Email address to contact the group",
      "default": null
    }
  ],
  "Aspect": {
    "name": "corpGroupEditableInfo"
  }
}